#include <Python.h>
#include <arrayobject.h>
#include <assert.h>
#include <time.h>
#include <matrix.h>

#include <queue.h>
#include <WorkRequest.h>
#include <WorkResult.h>

#include "pyConvRBM.cuh"
#include "ConvRBM.cuh"

static ConvRBM* model = NULL;

/* ==== Set up the methods table ====================== */
static PyMethodDef _ModelMethods[] =   {{ "initModel",      initModel,      METH_VARARGS },
                                        { "startBatch",     startBatch,     METH_VARARGS },
                                        { "finishBatch",    finishBatch,    METH_VARARGS },
                                        { "syncWithHost",   syncWithHost,   METH_VARARGS },
                                        { NULL, NULL } /* Sentinel - marks the end of this structure */
};

/*
 * Module initialization function. Required for python C extensions (and called automatically).
 */
void init_convrbm() {
    (void) Py_InitModule("_convrbm", _ModelMethods);
    import_array(); // Must be present for NumPy.  Called first after above line.
}

PyObject* initModel(PyObject *self, PyObject *args) {
    assert(model == NULL);

    PyArrayObject *hidVis, *visBiases, *hidBiases;
    PyArrayObject *hidVisInc, *visBiasInc, *hidBiasInc;
    PyArrayObject *hidMeans;
    PyArrayObject *visHidGlob, *hidBiasesGlob, *visHidIncGlob, *hidBiasIncGlob;
    PyArrayObject* hidMeansGlob;
    PyArrayObject* negDataPCD;

    int fixedBorderSize, localMaxSize, zeroPadding;
    double epsHidVis, epsVisBias, epsHidBias;
    double epsHidVisGlob, epsHidBiasGlob;
    double mom, weightCost;
    double sparseCost, sparseTarget;
    double sparseCostGlob, sparseTargetGlob;

    int minibatchSize, cd;
    int deviceID;

    if (!PyArg_ParseTuple(args, "O!O!O!O!O!O!O!O!O!O!O!O!O!iiidddddddddddiii",
                          &PyArray_Type, &hidVis,
                          &PyArray_Type, &visBiases,
                          &PyArray_Type, &hidBiases,
                          &PyArray_Type, &hidVisInc,
                          &PyArray_Type, &visBiasInc,
                          &PyArray_Type, &hidBiasInc,
                          &PyArray_Type, &hidMeans,
                          &PyArray_Type, &visHidGlob,
                          &PyArray_Type, &hidBiasesGlob,
                          &PyArray_Type, &visHidIncGlob,
                          &PyArray_Type, &hidBiasIncGlob,
                          &PyArray_Type, &hidMeansGlob,
                          &PyArray_Type, &negDataPCD,
                          &fixedBorderSize, &localMaxSize, &zeroPadding,
                          &epsHidVis, &epsVisBias, &epsHidBias,
                          &epsHidVisGlob, &epsHidBiasGlob,
                          &mom, &weightCost,
                          &sparseCost, &sparseTarget,
                          &sparseCostGlob, &sparseTargetGlob,
                          &minibatchSize, &cd,
                          &deviceID)) {
        return NULL;
    }

    model   = new ConvRBM(new Matrix(hidVis),
                          new Matrix(visBiases),
                          new Matrix(hidBiases),
                          new Matrix(hidVisInc),
                          new Matrix(visBiasInc),
                          new Matrix(hidBiasInc),
                          new Matrix(hidMeans),
                          new Matrix(visHidGlob),
                          new Matrix(hidBiasesGlob),
                          new Matrix(visHidIncGlob),
                          new Matrix(hidBiasIncGlob),
                          new Matrix(hidMeansGlob),
                          new Matrix(negDataPCD),
                          fixedBorderSize, localMaxSize, zeroPadding,
                          epsHidVis, epsVisBias, epsHidBias,
                          epsHidVisGlob, epsHidBiasGlob,
                          mom, weightCost,
                          sparseCost, sparseTarget,
                          sparseCostGlob, sparseTargetGlob,
                          minibatchSize, cd,
                          deviceID);

    model->start();
    return Py_BuildValue("i", 0);
}

/*
 * Starts training/testing on the given batch (asynchronous -- returns immediately).
 */
PyObject* startBatch(PyObject *self, PyObject *args) {
    assert(model != NULL);
    PyArrayObject* data;
    int test = 0;
    if (!PyArg_ParseTuple(args, "O!|i",
                          &PyArray_Type, &data,
                          &test)) {
        return NULL;
    }

    model->getWorkRequestQueue()->enqueue(new WorkRequest(test ? WorkRequest::TEST : WorkRequest::TRAIN, new Matrix(data)));
    return Py_BuildValue("i", 0);
}

/*
 * Waits for the trainer to finish training on the batch given to startBatch.
 * Returns average reconstruction error.
 */
PyObject* finishBatch(PyObject *self, PyObject *args) {
    assert(model != NULL);
    BatchDoneResult* res = dynamic_cast<BatchDoneResult*>(model->getWorkResultQueue()->dequeue());

    double batchError = res->getBatchError();
    delete res;
    return Py_BuildValue("(d)", batchError);
}

/*
 * Copies weight matrices from GPU to system memory.
 */
PyObject* syncWithHost(PyObject *self, PyObject *args) {
    assert(model != NULL);
    model->getWorkRequestQueue()->enqueue(new WorkRequest(WorkRequest::SYNC));
    SyncDoneResult* res = dynamic_cast<SyncDoneResult*>(model->getWorkResultQueue()->dequeue());
    delete res;
    return Py_BuildValue("i", 0);
}
